//#include <Wire.h> //already included in other libraries.
#include <Adafruit_MLX90614.h> // GY-906 library
#include <CircularBuffer.h> //MAX30100 related library
#include <MAX30100.h>
#include <MAX30100_BeatDetector.h>
#include <MAX30100_Filters.h>
#include <MAX30100_PulseOximeter.h>
#include <MAX30100_Registers.h>
#include <MAX30100_SpO2Calculator.h>

// MACRO Definitions
#define BP_START_PIN (2)         // start button of the blood pressure monitor device. connected to transistor base.
#define VALVE_PIN (3)            // checks if the blood pressure measurement is done.             

// Global variabls

//-------------------Blood Pressure Monitor variables
volatile byte i2c_data_rx;       // indicates there are available data from the i2c bus.
volatile uint16_t count;         // indicates the total number of data collected.
volatile uint8_t sys, dia, hr;   // stored the measure values: systolic, diastolic and heart rate.
//------------------------------------------------
//------------------------------------------Temp. variables
double amb, tar = 0; // current ambient and target temp.
//----------------------------------------------
//--------------------------------------------Oximeter variables
double har = 0; // current heart rate
int    spo2 = 0;// current oxygen level in blood
double lhar = 0;// last heart rate
int    lspo2 = 0; // oxygen level in blood
//-------------------------------------------------------
//-------------------serial variable
String Incoming_value;               //Variable for storing Incoming Value from HC-05
//------------------------------------------

//------------------------------------------.....
PulseOximeter pox;  //Oximeter object
Adafruit_MLX90614 mlx = Adafruit_MLX90614(); // mlx declaration for GY-906
//------------------------------------------

//------------------------ Function declaration
void onBeatDetected();
void printData();
void mesureTemp();
void mesureOxi();
void mesureBP();
void receiveEvent(int iData);
//---------------------------------------------

//==========================================================================================setup()
void setup() {
  // Initialize serial protocol for HC-05
  Serial.begin(115200);

  // pins configuration
  pinMode(BP_START_PIN, OUTPUT);
  pinMode(VALVE_PIN, INPUT);


  if (!pox.begin()) { // Initialize the oximeter MAX30100
    // Serial.println("oximeter connection FAILED");
    //for (;;);

  } else {
    // Serial.println("oximeter connection SUCCESS");

  }

  mlx.begin(); // Initialize the temp. sensor GY-906

}
//==========================================================================================

//==========================================================================================loop()
void loop() {
  if (Serial.available() > 0) { // receive serial data
    Incoming_value = Serial.readString();
  }
  //Serial.println(Incoming_value); // for debugging

  //--------------------------- receiving measure all command via Bluetooth ("sao")
  if (Incoming_value == "sao") {

    mesureTemp();

    pox.begin(); // reinitiate and set the proper I2C speed for the MAX30100
    while (((spo2 == 0) or (har == 0)) or ((har == lhar) and (spo2 == lspo2))) { // keep reading MAX30100 values until you get new non-zero ones

      for ( int i = 0; i < 600; i++) {
        mesureOxi();
      }
    }

    lspo2 = spo2;
    lhar = har;

    printData();

    mesureBP();

    Incoming_value = "0";
    Serial.print('\n');
  }

  //-------------------------------------

  //-------------------------------receiving monitor command via Bluetooth ("mon")
  while (Incoming_value == "mon") {

    while (Incoming_value != "close") {

      if (Incoming_value != "BP") {

        if (Serial.available() > 0) {
          Incoming_value =  Serial.readString();
        }

        mesureTemp();

        pox.begin(); // reinitiate and set the proper I2C speed for the MAX30100

        while (((spo2 == 0) or (har == 0)) or ((har == lhar) and (spo2 == lspo2)) ) { // keep reading MAX30100 values until you get new non-zero ones
          for ( int i = 0; i < 600; i++) {
            mesureOxi();
          }
        }

        if (Serial.available() > 0) {
          Incoming_value =  Serial.readString();
        }

        lspo2 = spo2;
        lhar = har;
        printData();
        Serial.print('\n');
      }

      if ( Incoming_value == "BP") {
        mesureBP();
        Serial.print('\n');
        Incoming_value = "0";
      }

      if (Serial.available() > 0) {
        Incoming_value =  Serial.readString();
      }

    }
  }
  //-----------------------------------
}
//==========================================================================================

/********************************************************************
  Temperature measurement:
  This function stores the value of the
  current ambient temperature in the global variable amb, and stores
  the current target temperature in the global variable tar.
********************************************************************/
void mesureTemp() {

  Wire.setClock(100000UL); // set the proper I2C speed for the GY-906
  amb = mlx.readAmbientTempC();
  tar = mlx.readObjectTempC();
}
/*******************************************************************/

/********************************************************************
   Oxygen and Heart Rate measurement:
   This function stores the value
   of the current Oxygen level in the global variable spo2, and
   stores the current Heart Rate in in the global variable har.
********************************************************************/
void mesureOxi() {
  pox.update(); //
  har = pox.getHeartRate();
  spo2 = pox.getSpO2();
}
/*******************************************************************/

/********************************************************************
  Sending data via Bluetooth: This function prints data in a
  specific format that is understandable by the Python app on the
  Raspberry Pi.
********************************************************************/
void printData() {

  Serial.print("Ambient,");
  Serial.print(amb);
  Serial.print(",C");
  Serial.print(",Target,");
  Serial.print(amb);
  Serial.print(",C,");
  Serial.print("heartreat,");
  Serial.print(har);
  Serial.print(",bpm,");
  Serial.print("SpO2,");
  Serial.print(spo2);
  Serial.print(",%,");

}
/*******************************************************************/

/*******************************************************************************
  Measuring Blood Pressure:
  This function starts by changing the I2C mode to remote mode
  holding the same address of the EEPROM in the Blood Pressure Monitor. Then,it
  mimics the pressing of the start button.
  After that, it delays for sometime to allow the valve to close. When it
  confirms that the valve is closed, it halts until it's released.
  During that halt, the I2C interrupt handler is triggered ( see the I2C
  interrupt handler description).
  After the valve release, it mimics the pressing of the start button again to
  turn off the device.
  Then, data is printed on the serial line to be sent over Bluetooth.
  Lastly, it puts the I2C mode back to Master, and reinitiate the GY-906 and
  MAX30100.
*******************************************************************************/
void mesureBP() {

  Wire.begin(0x50);                           // the address of the EEPROM is 0x50. The arduino should be the same.
  Wire.onReceive(receiveEvent);               // this is the interrupt initialization for the i2c data.
  digitalWrite(BP_START_PIN, HIGH);           // Emulating a push on the start button to start measurement.
  delay(200);
  digitalWrite(BP_START_PIN, LOW);
  //Serial.println(count);
  delay(2000);
  //Serial.println("Start BP check...");
  delay(2000);                          //need to secure that the value is already closed.
  count = 0;
  while (digitalRead(VALVE_PIN) == 0) // halts until valve is released.
  {
    //Serial.println("BP On-going..."); // for debugging porpos
    delay(1000);
    //Serial.println(count); // for debugging porpos
  }

  delay(2000);

  digitalWrite(BP_START_PIN, HIGH); // Emulating a push on the start button to turn off the device.
  delay(200);
  digitalWrite(BP_START_PIN, LOW);

  delay(500);
  //Serial.println(count); // for debugging porpos
  if (count == 0) // if no data was received then print Error.
  {
    Serial.print("BP,");
    Serial.print("Error");
    Serial.print("/");
    Serial.print("Error");
    Serial.print(",heartreat,");
    Serial.print("Error");
    Serial.print(",");
  }
  else if (count > 5) // if more than 5 data was sent on I2C then print values.
  {
    Serial.print("BP,");
    Serial.print(sys);
    Serial.print("/");
    Serial.print(dia);
    Serial.print(",heartreat,");
    Serial.print(hr);
    Serial.print(",");
    //Serial.println(count); // for debugging porpos.
  }


  count = 0; // reset I2C data Counter.
  Wire.begin();
}
/***************************************************************
  I2C interrupt handler: records all communication data
  between the EEPROM and the Blood Pressure Monitors MCU, and
  store the needed data in their corresponding global variable.
 **************************************************************/

void receiveEvent(int iData)   // Interrupt service routine.
{
  if ( iData > 0 )
  {
    while ( iData-- )
    {
      i2c_data_rx = Wire.read();
      // Serial.println(i2c_data_rx,HEX); // used for sniffing I2C data to locate needed information
      count++; // counts the total number of data collected.

      if (count == 5) // after sniffing, it was found that the systolic value is the 5th transmitted data
      {
        sys = i2c_data_rx;
      }
      if (count == 3) // after sniffing, it was found that the diastolic value is the 3rd transmitted data
      {
        dia = i2c_data_rx;
      }
      if (count == 4) // after sniffing, it was found that the Heart rate value is the 4th transmitted data
      {
        hr = i2c_data_rx;
      }
    }
  }
}
